# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import datetime
import json
import subprocess
import textwrap
from typing import IO, Any, Optional

from mkosi.config import ManifestFormat, OutputFormat
from mkosi.context import Context
from mkosi.distribution import PackageType
from mkosi.installer.apt import Apt
from mkosi.installer.pacman import Pacman
from mkosi.log import complete_step
from mkosi.run import run
from mkosi.util import read_env_file


@dataclasses.dataclass
class PackageManifest:
    """A description of a package

    The fields used here must match
    https://systemd.io/COREDUMP_PACKAGE_METADATA/#well-known-keys.
    """

    type: str
    name: str
    version: str
    architecture: str
    size: int

    def as_dict(self) -> dict[str, str]:
        return {
            "type": self.type,
            "name": self.name,
            "version": self.version,
            "architecture": self.architecture,
        }


@dataclasses.dataclass
class SourcePackageManifest:
    name: str
    changelog: Optional[str]
    packages: list[PackageManifest] = dataclasses.field(default_factory=list)

    def add(self, package: PackageManifest) -> None:
        self.packages.append(package)

    def report(self) -> str:
        size = sum(p.size for p in self.packages)

        t = textwrap.dedent(
            f"""\
            SourcePackage: {self.name}
            Packages:      {" ".join(p.name for p in self.packages)}
            Size:          {size}
            """
        )
        if self.changelog:
            t += f"""\nChangelog:\n{self.changelog}\n"""
        return t


@dataclasses.dataclass
class Manifest:
    context: Context
    packages: list[PackageManifest] = dataclasses.field(default_factory=list)
    source_packages: dict[str, SourcePackageManifest] = dataclasses.field(default_factory=dict)
    extension: dict[str, str] = dataclasses.field(default_factory=dict)

    _init_timestamp: datetime.datetime = dataclasses.field(
        init=False, default_factory=lambda: datetime.datetime.now().replace(microsecond=0)
    )

    def need_source_info(self) -> bool:
        return ManifestFormat.changelog in self.context.config.manifest_format

    def record_packages(self) -> None:
        with complete_step("Recording packages in manifest…"):
            if self.context.config.distribution.installer.package_type() == PackageType.rpm:
                self.record_rpm_packages()
            if self.context.config.distribution.installer.package_type() == PackageType.deb:
                self.record_deb_packages()
            if self.context.config.distribution.installer.package_type() == PackageType.pkg:
                self.record_pkg_packages()

    def record_rpm_packages(self) -> None:
        c = run(
            [
                "rpm",
                "--root=/buildroot",
                "--query",
                "--all",
                "--queryformat", r"%{NEVRA}\t%{SOURCERPM}\t%{NAME}\t%{ARCH}\t%{LONGSIZE}\t%{INSTALLTIME}\n",
            ],
            stdout=subprocess.PIPE,
            sandbox=(
                self.context.sandbox(options=["--ro-bind", self.context.root, "/buildroot"])
            ),
        )  # fmt: skip

        packages = sorted(c.stdout.splitlines())

        for package in packages:
            nevra, srpm, name, arch, size, installtime = package.split("\t")

            assert nevra.startswith(f"{name}-")
            evra = nevra.removeprefix(f"{name}-")
            # Some packages have architecture '(none)', and it's not part of NEVRA, e.g.:
            # gpg-pubkey-45719a39-5f2c0192 gpg-pubkey (none) 0 1635985199
            if arch != "(none)":
                assert nevra.endswith(f".{arch}")
                evr = evra.removesuffix(f".{arch}")
            else:
                evr = evra
                arch = ""

            # If we are creating a layer based on a BaseImage=, e.g. a sysext, filter by
            # packages that were installed in this execution of mkosi. We assume that the
            # upper layer is put together in one go, which currently is always true.
            if (
                self.context.config.base_trees
                and datetime.datetime.fromtimestamp(int(installtime)) < self._init_timestamp
            ):
                continue

            manifest = PackageManifest("rpm", name, evr, arch, int(size))
            self.packages.append(manifest)

            if not self.need_source_info():
                continue

            source = self.source_packages.get(srpm)
            if source is None:
                c = run(
                    [
                        "rpm",
                        "--root=/buildroot",
                        "--query",
                        "--changelog",
                        nevra,
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                    sandbox=self.context.sandbox(options=["--ro-bind", self.context.root, "/buildroot"]),
                )
                changelog = c.stdout.strip()
                source = SourcePackageManifest(srpm, changelog)
                self.source_packages[srpm] = source

            source.add(manifest)

    def record_deb_packages(self) -> None:
        c = run(
            [
                "dpkg-query",
                "--admindir=/buildroot/var/lib/dpkg",
                "--show",
                "--showformat", r"${Package}\t${source:Package}\t${Version}\t${Architecture}\t${Installed-Size}\t${db-fsys:Last-Modified}\n",  # noqa: E501
            ],
            stdout=subprocess.PIPE,
            sandbox=self.context.sandbox(options=["--ro-bind", self.context.root, "/buildroot"]),
        )  # fmt: skip

        packages = sorted(c.stdout.splitlines())

        for package in packages:
            name, source, version, arch, size, installtime = package.split("\t")

            # dpkg records the size in KBs, the field is optional db-fsys:Last-Modified is not available in
            # very old dpkg, so just skip creating the manifest for sysext when building on very old
            # distributions by setting the timestamp to epoch. This only affects Ubuntu Bionic which is
            # nearing EOL.  If we are creating a layer based on a BaseImage=, e.g. a sysext, filter by
            # packages that were installed in this execution of mkosi. We assume that the upper layer is put
            # together in one go, which currently is always true.
            install_timestamp = datetime.datetime.fromtimestamp(int(installtime) if installtime else 0)
            if self.context.config.base_trees and install_timestamp < self._init_timestamp:
                continue

            manifest = PackageManifest("deb", name, version, arch, int(size or 0) * 1024)
            self.packages.append(manifest)

            if not self.need_source_info():
                continue

            source_package = self.source_packages.get(source)
            if source_package is None:
                # Yes, --quiet is specified twice, to avoid output about download stats. Note that the
                # argument of the 'changelog' verb is the binary package name, not the source package
                # name. We also have to set "Dir" explicitly because apt has no separate option to configure
                # the changelog directory. Apt.invoke() sets all options that are interpreted relative to Dir
                # to absolute paths by default so this is safe.
                result = Apt.invoke(
                    self.context,
                    "changelog",
                    ["--quiet", "--quiet", "-o", "Dir=/buildroot", name],
                    stdout=subprocess.PIPE,
                )
                source_package = SourcePackageManifest(source, result.stdout.strip())
                self.source_packages[source] = source_package

            source_package.add(manifest)

    def record_pkg_packages(self) -> None:
        packages = sorted((self.context.root / "var/lib/pacman/local").glob("*/desc"))

        for desc in packages:
            name, version, source, arch = Pacman.parse_pkg_desc(desc)
            package = PackageManifest("pkg", name, version, arch, 0)
            self.packages.append(package)

            source_package = self.source_packages.get(source)
            if source_package is None:
                source_package = SourcePackageManifest(source, None)
                self.source_packages[source] = source_package
            source_package.add(package)

    def record_extension_release(self) -> None:
        if self.context.config.output_format not in (OutputFormat.sysext, OutputFormat.confext):
            return

        with complete_step(f"Recording {type} information in manifest…"):
            d = "usr/lib" if self.context.config.output_format == OutputFormat.sysext else "etc"
            p = self.context.root / d / f"extension-release.d/extension-release.{self.context.config.output}"

            self.extension = read_env_file(p)

    def has_data(self) -> bool:
        # We might add more data in the future
        return len(self.packages) > 0 or len(self.extension) > 0

    def as_dict(self) -> dict[str, Any]:
        config = {
            "name": self.context.config.image_id or "image",
            "distribution": str(self.context.config.distribution),
            "architecture": str(self.context.config.architecture),
            "output_format": str(self.context.config.output_format),
        }
        if self.context.config.image_version is not None:
            config["version"] = self.context.config.image_version
        if self.context.config.release is not None:
            config["release"] = self.context.config.release

        return {
            # Bump this when incompatible changes are made to the manifest format.
            "manifest_version": 1,
            # Describe the image itself.
            "config": config,
            # Describe the image content in terms of packages.
            "packages": [package.as_dict() for package in self.packages],
            # Describe the SYSEXT / CONFEXT metadata.
            "extension": self.extension,
        }

    def write_json(self, out: IO[str]) -> None:
        json.dump(self.as_dict(), out, indent=2)

    def write_package_report(self, out: IO[str]) -> None:
        """Create a human-readable report about packages

        This is modelled after "Fedora compose reports" that are sent
        to fedora-devel. The format describes added and removed
        packages, and includes the changelogs. A diff between two such
        reports shows what changed *in* the packages quite nicely.
        """
        out.write(f"Packages: {len(self.packages)}\n")
        out.write(f"Size:     {sum(p.size for p in self.packages)}")

        for package in self.source_packages.values():
            out.write(f"\n{80 * '-'}\n")
            out.write(package.report())



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Sequence
from contextlib import AbstractContextManager
from pathlib import Path

from mkosi.config import Config, ConfigFeature, OutputFormat
from mkosi.context import Context
from mkosi.mounts import finalize_certificate_mounts
from mkosi.run import apivfs_options, finalize_interpreter, finalize_passwd_symlinks, find_binary
from mkosi.tree import rmtree
from mkosi.util import PathString, flatten, startswith


class PackageManager:
    @classmethod
    def executable(cls, config: Config) -> str:
        return "custom"

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("custom")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        return []

    @classmethod
    def package_globs(cls) -> list[str]:
        return []

    @classmethod
    def state_subdirs(cls) -> list[Path]:
        return []

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        return {}

    @classmethod
    def architecture(cls, context: Context) -> str:
        return context.config.distribution.installer.architecture(context.config.architecture)

    @classmethod
    def finalize_environment(cls, context: Context) -> dict[str, str]:
        env = {
            "HOME": "/",  # Make sure rpm doesn't pick up ~/.rpmmacros and ~/.rpmrc.
            # systemd's chroot detection doesn't work when unprivileged so tell it explicitly.
            "SYSTEMD_IN_CHROOT": "1",
        }

        if "SYSTEMD_HWDB_UPDATE_BYPASS" not in context.config.finalize_environment():
            env["SYSTEMD_HWDB_UPDATE_BYPASS"] = "1"

        if (
            "KERNEL_INSTALL_BYPASS" not in context.config.finalize_environment()
            and context.config.bootable != ConfigFeature.disabled
        ):
            env["KERNEL_INSTALL_BYPASS"] = "1"
        else:
            env |= {
                "BOOT_ROOT": "/boot",
                # Required to make 90-loaderentry.install put the right paths into the bootloader entry.
                "BOOT_MNT": "/boot",
                # Hack to tell dracut to not create a hostonly initrd when it's invoked by kernel-install.
                "hostonly_l": "no",
            }

        return context.config.finalize_environment() | env

    @classmethod
    def env_cmd(cls, context: Context) -> list[PathString]:
        return ["env", *([f"{k}={v}" for k, v in cls.finalize_environment(context).items()])]

    @classmethod
    def mounts(cls, context: Context) -> list[PathString]:
        mounts = [
            *finalize_certificate_mounts(context.config),
            "--bind", context.repository, "/repository",
        ]  # fmt: skip

        if context.config.local_mirror and (mirror := startswith(context.config.local_mirror, "file://")):
            mounts += ["--ro-bind", mirror, mirror]

        subdir = context.config.distribution.installer.package_manager(context.config).subdir(context.config)

        src = context.metadata_dir / "lib" / subdir
        mounts += flatten(
            ("--bind", src / state_subdir, Path("/var/lib") / subdir / state_subdir)
            for state_subdir in context.config.distribution.installer.package_manager(
                context.config
            ).state_subdirs()
        )

        src = context.metadata_dir / "cache" / subdir
        caches = context.config.distribution.installer.package_manager(context.config).package_subdirs(src)

        # If there are no package cache subdirectories, we always operate on the package cache directory,
        # since we can't do any mount tricks to combine caches from different locations in this case.
        if caches == [(Path("."), Path("."))]:
            mounts += [
                "--bind",
                context.config.package_cache_dir_or_default() / "cache" / subdir,
                Path("/var/cache") / subdir,
            ]
        else:
            mounts += ["--bind", src, Path("/var/cache") / subdir]

            # If we're not operating on the configured package cache directory, we're operating on a snapshot
            # of the repository metadata. To make sure any downloaded packages are still cached in the
            # configured package cache directory in this scenario, we mount in the relevant directories from
            # the configured package cache directory.
            if context.metadata_dir != context.config.package_cache_dir_or_default():
                mounts += flatten(
                    (
                        "--bind",
                        context.config.package_cache_dir_or_default() / "cache" / subdir / srcsubdir,
                        Path("/var/cache") / subdir / dstsubdir,
                    )
                    for srcsubdir, dstsubdir in caches
                    if (
                        context.config.package_cache_dir_or_default() / "cache" / subdir / srcsubdir
                    ).exists()
                )

        return mounts

    @classmethod
    def options(cls, *, root: PathString, apivfs: bool = True) -> list[PathString]:
        return [
            *(apivfs_options() if apivfs else []),
            "--become-root",
            "--suppress-chown",
            "--suppress-sync",
            # Make sure /etc/machine-id is not overwritten by any package manager post install scripts.
            "--ro-bind-try", Path(root) / "etc/machine-id", "/buildroot/etc/machine-id",
            # Nudge gpg to create its sockets in /run by making sure /run/user/0 exists.
            "--dir", "/run/user/0",
            # Some package managers (e.g. dpkg) read from the host's /etc/passwd instead of the buildroot's
            # /etc/passwd so we symlink /etc/passwd from the buildroot to make sure it gets used.
            *(finalize_passwd_symlinks("/buildroot") if apivfs else []),
        ]  # fmt: skip

    @classmethod
    def apivfs_script_cmd(cls, context: Context) -> list[PathString]:
        return [
            finalize_interpreter(bool(context.config.tools_tree)), "-SI", "/sandbox.py",
            "--bind", "/", "/",
            "--same-dir",
            "--bind", "/var/tmp", "/buildroot/var/tmp",
            *apivfs_options(),
            *cls.options(root="/buildroot"),
        ]  # fmt: skip

    @classmethod
    def sandbox(
        cls,
        context: Context,
        *,
        apivfs: bool,
        options: Sequence[PathString] = (),
    ) -> AbstractContextManager[list[PathString]]:
        return context.sandbox(
            network=True,
            options=[
                *context.rootoptions(),
                *cls.mounts(context),
                *cls.options(root=context.root, apivfs=apivfs),
                *options,
            ],
        )  # fmt: skip

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        pass

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        pass

    @classmethod
    def sync(cls, context: Context, force: bool) -> None:
        pass

    @classmethod
    def createrepo(cls, context: Context) -> None:
        pass


def clean_package_manager_metadata(context: Context) -> None:
    """
    Remove package manager metadata

    Try them all regardless of the distro: metadata is only removed if
    the package manager is not present in the image.
    """
    subdir = context.config.distribution.installer.package_manager(context.config).subdir(context.config)

    if context.config.clean_package_metadata == ConfigFeature.disabled:
        return

    if context.config.clean_package_metadata == ConfigFeature.auto and context.config.output_format in (
        OutputFormat.directory,
        OutputFormat.tar,
    ):
        return

    # If cleaning is not explicitly requested, keep the repository metadata if we're building a directory or
    # tar image (which are often used as a base tree for extension images and thus should retain package
    # manager metadata) or if the corresponding package manager is installed in the image.

    executable = context.config.distribution.installer.package_manager(context.config).executable(
        context.config
    )
    remove = []

    for tool, paths in (
        ("rpm",      ["var/lib/rpm", "usr/lib/sysimage/rpm"]),
        ("dnf5",     ["usr/lib/sysimage/libdnf5"]),
        ("dpkg",     ["var/lib/dpkg"]),
        (executable, [f"var/lib/{subdir}", f"var/cache/{subdir}"]),
    ):  # fmt: skip
        if context.config.clean_package_metadata == ConfigFeature.enabled or not find_binary(
            tool, root=context.root
        ):
            remove += [context.root / p for p in paths if (context.root / p).exists()]

    rmtree(*remove, sandbox=context.sandbox)



# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os
import stat
import tempfile
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Optional, Union

from mkosi.config import BuildSourcesEphemeral, Config
from mkosi.log import die
from mkosi.sandbox import OVERLAYFS_SUPER_MAGIC, OverlayOperation, statfs
from mkosi.util import PathString, flatten


def stat_is_whiteout(st: os.stat_result) -> bool:
    return stat.S_ISCHR(st.st_mode) and st.st_rdev == 0


def delete_whiteout_files(path: Path) -> None:
    """Delete any char(0,0) device nodes underneath @path

    Overlayfs uses such files to mark "whiteouts" (files present in
    the lower layers, but removed in the upper one).
    """
    for entry in path.rglob("*"):
        # TODO: Use Path.stat() once we depend on Python 3.10+.
        if stat_is_whiteout(os.stat(entry, follow_symlinks=False)):
            entry.unlink()


def finalize_volatile_tmpdir() -> Path:
    if tempfile.tempdir and statfs(tempfile.tempdir) != OVERLAYFS_SUPER_MAGIC:
        return Path(tempfile.tempdir)

    for path in ("/var/tmp", "/tmp", "/dev/shm"):
        if Path(path).exists() and statfs(path) != OVERLAYFS_SUPER_MAGIC:
            return Path(path)

    die(
        "Cannot find temporary directory for volatile overlayfs upperdir",
        hint="Either /var/tmp, /tmp or /dev/shm must exist and not be located on overlayfs",
    )


@contextlib.contextmanager
def mount_overlay(
    lowerdirs: Sequence[Path],
    dst: Path,
    *,
    upperdir: Optional[Path] = None,
) -> Iterator[Path]:
    with contextlib.ExitStack() as stack:
        if upperdir is None:
            upperdir = Path(
                stack.enter_context(
                    tempfile.TemporaryDirectory(
                        prefix="volatile-overlay.",
                        dir=finalize_volatile_tmpdir(),
                    )
                )
            )
            st = lowerdirs[-1].stat()
            os.chmod(upperdir, st.st_mode)

        workdir = Path(
            stack.enter_context(
                tempfile.TemporaryDirectory(dir=upperdir.parent, prefix=f"{upperdir.name}-workdir")
            )
        )

        try:
            with OverlayOperation(
                tuple(os.fspath(p) for p in lowerdirs),
                os.fspath(upperdir),
                os.fspath(workdir),
                os.fspath(dst),
            ):
                yield dst
        finally:
            delete_whiteout_files(upperdir)


@contextlib.contextmanager
def finalize_source_mounts(
    config: Config,
    *,
    ephemeral: Union[BuildSourcesEphemeral, bool],
) -> Iterator[list[PathString]]:
    with contextlib.ExitStack() as stack:
        options: list[PathString] = []

        for t in config.build_sources:
            src, dst = t.with_prefix("/work/src")

            if ephemeral:
                if ephemeral == BuildSourcesEphemeral.buildcache:
                    if config.build_dir is None:
                        die(
                            "BuildSourcesEphemeral=buildcache was configured, but no build directory exists.",  # noqa: E501
                            hint="Configure BuildDirectory= or create mkosi.builddir.",
                        )

                    upperdir = config.build_subdir / f"mkosi.buildovl.{src.name}"
                    upperdir.mkdir(mode=src.stat().st_mode, exist_ok=True)
                else:
                    upperdir = Path(
                        stack.enter_context(
                            tempfile.TemporaryDirectory(
                                prefix="volatile-overlay.",
                                dir=finalize_volatile_tmpdir(),
                            )
                        )
                    )
                    os.chmod(upperdir, src.stat().st_mode)

                workdir = Path(
                    stack.enter_context(
                        tempfile.TemporaryDirectory(dir=upperdir.parent, prefix=f"{upperdir.name}-workdir.")
                    )
                )

                options += [
                    "--overlay-lowerdir", src,
                    "--overlay-upperdir", upperdir,
                    "--overlay-workdir", workdir,
                    "--overlay", dst,
                ]  # fmt: skip
            else:
                options += ["--bind", src, dst]

        yield options


def finalize_certificate_mounts(config: Config, relaxed: bool = False) -> list[PathString]:
    mounts = []
    root = config.tools() if config.tools_tree_certificates else Path("/")

    if not relaxed or root != Path("/"):
        subdirs = [Path("var/lib/ca-certificates")]
        if not relaxed:
            subdirs += [
                Path("etc/pki/ca-trust"),
                Path("etc/pki/tls"),
                Path("etc/ssl"),
                Path("etc/ca-certificates"),
            ]

        mounts += [
            (root / subdir, Path("/") / subdir)
            for subdir in subdirs
            if (root / subdir).exists() and any(p for p in (root / subdir).rglob("*") if not p.is_dir())
        ]

    return flatten(("--ro-bind", src, target) for src, target in sorted(set(mounts), key=lambda s: s[1]))



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable

from mkosi.config import Architecture
from mkosi.context import Context
from mkosi.distribution import Distribution, fedora, join_mirror
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey
from mkosi.log import die


class Installer(fedora.Installer, distribution=Distribution.mageia):
    @classmethod
    def pretty_name(cls) -> str:
        return "Mageia"

    @classmethod
    def filesystem(cls) -> str:
        return "ext4"

    @classmethod
    def default_release(cls) -> str:
        return "cauldron"

    @classmethod
    def install(cls, context: Context) -> None:
        cls.install_packages(context, ["filesystem"], apivfs=False)

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.snapshot:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        gpgurls = (
            find_rpm_gpgkey(
                context,
                "RPM-GPG-KEY-Mageia",
                "https://mirrors.kernel.org/mageia/distrib/$releasever/$basearch/media/core/release/media_info/pubkey",
            ),
        )

        if context.config.local_mirror:
            yield RpmRepository("core-release", f"baseurl={context.config.local_mirror}", gpgurls)
            return

        if context.config.mirror:
            url = (
                f"baseurl={join_mirror(context.config.mirror, 'distrib/$releasever/$basearch/media/core/')}"
            )
            yield RpmRepository("core-release", f"{url}/release", gpgurls)
            yield RpmRepository("core-updates", f"{url}/updates/", gpgurls)
        else:
            url = "mirrorlist=https://www.mageia.org/mirrorlist/?release=$releasever&arch=$basearch&section=core"
            yield RpmRepository("core-release", f"{url}&repo=release", gpgurls)
            yield RpmRepository("core-updates", f"{url}&repo=updates", gpgurls)

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.x86_64: "x86_64",
            Architecture.arm64:  "aarch64",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by Mageia")

        return a



# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import dataclasses
import enum
import io
import shlex
from collections.abc import Iterable, Mapping
from pathlib import Path
from textwrap import indent
from typing import Optional, Union

from mkosi import config
from mkosi.log import die
from mkosi.util import StrEnum


class CompGen(StrEnum):
    default = enum.auto()
    files = enum.auto()
    dirs = enum.auto()

    @staticmethod
    def from_action(action: argparse.Action) -> "CompGen":
        if isinstance(action.default, Path):
            if action.default.is_dir():
                return CompGen.dirs
            else:
                return CompGen.files
        # TODO: the type of action.type is Union[Callable[[str], Any], FileType]
        # the type of Path is type, but Path also works in this position,
        # because the constructor is a callable from str -> Path
        elif action.type is not None and (isinstance(action.type, type) and issubclass(action.type, Path)):  # type: ignore
            if isinstance(action.default, Path) and action.default.is_dir():  # type: ignore
                return CompGen.dirs
            else:
                return CompGen.files

        return CompGen.default

    def to_bash(self) -> str:
        return f"_mkosi_compgen_{self}"

    def to_fish(self) -> str:
        if self == CompGen.files:
            return "--force-files"
        elif self == CompGen.dirs:
            return "--force-files -a '(__fish_complete_directories)'"
        else:
            return "-f"

    def to_zsh(self) -> str:
        if self == CompGen.files:
            return ":path:_files -/"
        elif self == CompGen.dirs:
            return ":directory:_files -f"
        else:
            return ""


@dataclasses.dataclass(frozen=True)
class CompletionItem:
    short: Optional[str]
    long: Optional[str]
    help: Optional[str]
    choices: list[str]
    compgen: CompGen


def collect_completion_arguments() -> list[CompletionItem]:
    parser = config.create_argument_parser()

    options = [
        CompletionItem(
            short=next((s for s in action.option_strings if not s.startswith("--")), None),
            long=next((s for s in action.option_strings if s.startswith("--")), None),
            help=action.help,
            choices=[str(c) for c in action.choices] if action.choices is not None else [],
            compgen=CompGen.from_action(action),
        )
        for action in parser._actions
        if (
            action.option_strings
            and action.help != argparse.SUPPRESS
            and action.dest not in config.SETTINGS_LOOKUP_BY_DEST
        )
    ]

    options += [
        CompletionItem(
            short=setting.short,
            long=setting.long,
            help=setting.help,
            choices=[str(c) for c in setting.choices] if setting.choices is not None else [],
            compgen=CompGen.default,
        )
        for setting in config.SETTINGS
    ]

    return options


def finalize_completion_bash(options: list[CompletionItem], resources: Path) -> str:
    def to_bash_array(name: str, entries: Iterable[str]) -> str:
        return f"{name.replace('-', '_')}=(" + " ".join(shlex.quote(str(e)) for e in entries) + ")"

    def to_bash_hasharray(name: str, entries: Mapping[str, Union[str, int]]) -> str:
        return (
            f"{name.replace('-', '_')}=("
            + " ".join(f"[{shlex.quote(str(k))}]={shlex.quote(str(v))}" for k, v in entries.items())
            + ")"
        )

    completion = resources / "completion.bash"

    options_by_key = {o.short: o for o in options if o.short} | {o.long: o for o in options if o.long}

    template = completion.read_text()
    with io.StringIO() as c:
        c.write(to_bash_array("_mkosi_options", options_by_key.keys()))
        c.write("\n\n")

        choices = to_bash_hasharray(
            "_mkosi_choices",
            {optname: " ".join(v.choices) for optname, v in options_by_key.items() if v.choices},
        )
        c.write(choices)
        c.write("\n\n")

        compgen = to_bash_hasharray(
            "_mkosi_compgen",
            {
                optname: v.compgen.to_bash()
                for optname, v in options_by_key.items()
                if v.compgen != CompGen.default
            },
        )
        c.write(compgen)
        c.write("\n\n")

        c.write(to_bash_hasharray("_mkosi_verbs", {str(v): str(v) for v in config.Verb}))

        definitions = c.getvalue()

    return template.replace("##VARIABLEDEFINITIONS##", indent(definitions, " " * 4))


def finalize_completion_fish(options: list[CompletionItem], resources: Path) -> str:
    with io.StringIO() as c:
        # REUSE-IgnoreStart
        c.write("# SPDX-License-Identifier: LGPL-2.1-or-later\n\n")
        # REUSE-IgnoreEnd
        c.write("complete -c mkosi -f\n")

        c.write("complete -c mkosi -n '__fish_is_first_token' -a \"")
        c.write(" ".join(str(v) for v in config.Verb))
        c.write('"\n')
        # Complete paths after first token
        c.write("complete -c mkosi -F -n 'not __fish_is_first_token'\n")

        for option in options:
            if not option.short and not option.long:
                continue

            c.write("complete -c mkosi ")
            if option.short:
                c.write(f"-s {option.short.lstrip('-')} ")
            if option.long:
                c.write(f"-l {option.long.lstrip('-')} ")
            c.write("-r ")
            if option.choices:
                c.write('-a "')
                c.write(" ".join(option.choices))
                c.write('" ')
            if option.help is not None:
                help = option.help.replace("'", "\\'")
                c.write(f'-d "{help}" ')
            c.write(option.compgen.to_fish())
            c.write("\n")

        return c.getvalue()


def finalize_completion_zsh(options: list[CompletionItem], resources: Path) -> str:
    def to_zsh_array(name: str, entries: Iterable[str]) -> str:
        return (
            f"declare -a {name.replace('-', '_')}=(" + " ".join(shlex.quote(str(e)) for e in entries) + ")"
        )

    completion = resources / "completion.zsh"

    with io.StringIO() as c:
        c.write(completion.read_text())
        c.write("\n")

        c.write(to_zsh_array("_mkosi_verbs", [str(v) for v in config.Verb]))
        c.write("\n\n")

        c.write("_arguments -s \\\n")
        c.write("    '(- *)'{-h,--help}'[Show this help]' \\\n")
        c.write("    '(- *)--version[Show package version]' \\\n")

        for option in options:
            if not option.short and not option.long:
                continue

            posix = option.help and "'" in option.help
            open_quote = "$'" if posix else "'"
            if option.short and option.long:
                c.write(f"    '({option.short} {option.long})'{{{option.short},{option.long}}}{open_quote}")
            else:
                c.write(f"    {open_quote}{option.short or option.long}")

            if option.help:
                help = option.help.replace("'", r"\'")
                c.write(f"[{help}]")
            if option.choices:
                # TODO: maybe use metavar here? At least for me it's not shown, though
                c.write(":arg:(")
                c.write(" ".join(option.choices))
                c.write(")")
            c.write(option.compgen.to_zsh())
            c.write("' \\\n")

        c.write("    '*::mkosi verb:_mkosi_verb'\n\n")

        return c.getvalue()


def print_completion(args: config.Args, *, resources: Path) -> None:
    if not args.cmdline:
        die(
            "No shell to generate completion script for specified",
            hint="Please specify either one of: bash, fish, zsh",
        )

    shell = args.cmdline[0]
    if shell == "bash":
        func = finalize_completion_bash
    elif shell == "fish":
        func = finalize_completion_fish
    elif shell == "zsh":
        func = finalize_completion_zsh
    else:
        die(
            f"{shell!r} is not supported for completion scripts.",
            hint="Please specify either one of: bash, fish, zsh",
        )

    completion_args = collect_completion_arguments()
    print(func(completion_args, resources))



# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
import subprocess
from pathlib import Path

from mkosi.config import DocFormat
from mkosi.log import die
from mkosi.pager import page
from mkosi.run import find_binary, run


def show_docs(
    manual: str,
    formats: list[DocFormat],
    *,
    man_chapter: int = 1,
    resources: Path,
    pager: bool = True,
) -> None:
    while formats:
        form = formats.pop(0)
        try:
            if form == DocFormat.man:
                man = resources / f"man/{manual}.{man_chapter}"
                if not man.exists():
                    raise FileNotFoundError()
                run(["man", "--local-file", man])
                return
            elif form == DocFormat.pandoc:
                if not find_binary("pandoc"):
                    logging.warn("pandoc is not available")
                    continue
                pandoc = run(
                    [
                        "pandoc",
                        "-t", "man",
                        "-s", resources / f"man/{manual}.{man_chapter}.md",
                        "--lua-filter", resources / "pandoc/md2man.lua",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                    log=False,
                )  # fmt: skip
                run(["man", "--local-file", "-"], input=pandoc.stdout)
                return
            elif form == DocFormat.markdown:
                page((resources / f"man/{manual}.{man_chapter}.md").read_text(), pager)
                return
            elif form == DocFormat.system:
                run(["man", str(man_chapter), manual], log=False)
                return
        except (FileNotFoundError, subprocess.CalledProcessError) as e:
            if not formats:
                if isinstance(e, FileNotFoundError):
                    die("The mkosi package does not contain the man page {manual!r}.")
                raise e



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable

from mkosi.config import Architecture, Config
from mkosi.context import Context
from mkosi.curl import curl
from mkosi.distribution import (
    Distribution,
    DistributionInstaller,
    PackageType,
    join_mirror,
)
from mkosi.installer.dnf import Dnf
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey, setup_rpm
from mkosi.log import die
from mkosi.util import startswith
from mkosi.versioncomp import GenericVersion

CENTOS_SIG_REPO_PRIORITY = 50


class Installer(DistributionInstaller, distribution=Distribution.centos):
    @classmethod
    def pretty_name(cls) -> str:
        return "CentOS"

    @classmethod
    def filesystem(cls) -> str:
        return "xfs"

    @classmethod
    def package_type(cls) -> PackageType:
        return PackageType.rpm

    @classmethod
    def default_release(cls) -> str:
        return "10"

    @classmethod
    def default_tools_tree_distribution(cls) -> Distribution:
        return Distribution.fedora

    @classmethod
    def major_release(cls, config: "Config") -> str:
        return config.release.partition(".")[0]

    @classmethod
    def package_manager(cls, config: "Config") -> type[Dnf]:
        return Dnf

    @classmethod
    def grub_prefix(cls) -> str:
        return "grub2"

    @classmethod
    def dbpath(cls, context: Context) -> str:
        # The Hyperscale SIG uses /usr/lib/sysimage/rpm in its rebuild of rpm for C9S that's shipped in the
        # hyperscale-packages-experimental repository.
        if (
            GenericVersion(context.config.release) > 9
            or "hyperscale-packages-experimental" in context.config.repositories
        ):
            return "/usr/lib/sysimage/rpm"

        return "/var/lib/rpm"

    @classmethod
    def setup(cls, context: Context) -> None:
        if GenericVersion(context.config.release) <= 8:
            die(f"{cls.pretty_name()} Stream 8 or earlier variants are not supported")

        setup_rpm(context, dbpath=cls.dbpath(context))

        Dnf.setup(context, list(cls.repositories(context)))
        (context.sandbox_tree / "etc/dnf/vars/stream").write_text(
            f"{cls.major_release(context.config)}-stream\n"
        )

    @classmethod
    def install(cls, context: Context) -> None:
        cls.install_packages(context, ["basesystem"], apivfs=False)

    @classmethod
    def architecture(cls, arch: Architecture) -> str:
        a = {
            Architecture.x86_64:   "x86_64",
            Architecture.ppc64_le: "ppc64le",
            Architecture.s390x:    "s390x",
            Architecture.arm64:    "aarch64",
        }.get(arch)  # fmt: skip

        if not a:
            die(f"Architecture {a} is not supported by {cls.pretty_name()}")

        return a

    @classmethod
    def gpgurls(cls, context: Context) -> tuple[str, ...]:
        # First, start with the names of the appropriate keys in /etc/pki/rpm-gpg.

        if GenericVersion(context.config.release) == 9:
            rel = "RPM-GPG-KEY-centosofficial"
        else:
            rel = "RPM-GPG-KEY-centosofficial-SHA256"

        one = find_rpm_gpgkey(context, rel, required=False)

        # Next, follow up with the names of the appropriate keys in /usr/share/distribution-gpg-keys.

        if GenericVersion(context.config.release) == 9:
            rel = "RPM-GPG-KEY-CentOS-Official"
        else:
            rel = "RPM-GPG-KEY-CentOS-Official-SHA256"

        # The key in /usr/share/distribution-gpg-keys is only required if we didn't find one in
        # /etc/pki/rpm-gpg.
        two = find_rpm_gpgkey(context, rel, f"https://www.centos.org/keys/{rel}", required=bool(one))

        # Finally, look up the key for the SIG-Extras repository.

        sig = find_rpm_gpgkey(
            context,
            "RPM-GPG-KEY-CentOS-SIG-Extras",
            "https://www.centos.org/keys/RPM-GPG-KEY-CentOS-SIG-Extras",
        )

        return tuple(key for key in (one, two, sig) if key is not None)

    @classmethod
    def repository_variants(
        cls,
        context: Context,
        gpgurls: tuple[str, ...],
        repo: str,
    ) -> Iterable[RpmRepository]:
        mirror = context.config.mirror
        if not mirror and context.config.snapshot:
            mirror = "https://composes.stream.centos.org"

        if context.config.snapshot and mirror not in (
            "https://composes.stream.centos.org",
            "https://mirror.facebook.net/centos-composes",
        ):
            die(
                f"Snapshot= is only supported for {cls.pretty_name()} if Mirror=https://composes.stream.centos.org"
            )

        if (
            mirror in ("https://composes.stream.centos.org", "https://mirror.facebook.net/centos-composes")
            and not context.config.snapshot
        ):
            die(f"Snapshot= must be used on {cls.pretty_name()} if Mirror={mirror}")

        if context.config.local_mirror:
            yield RpmRepository(repo, f"baseurl={context.config.local_mirror}", gpgurls)

        elif mirror:
            if mirror == "https://composes.stream.centos.org":
                subdir = f"stream-{context.config.release}/production"
            elif mirror == "https://mirror.facebook.net/centos-composes":
                subdir = context.config.release
            elif repo == "extras":
                subdir = "SIGs/$stream"
            else:
                subdir = "$stream"

            if context.config.snapshot:
                subdir += f"/CentOS-Stream-{context.config.release}-{context.config.snapshot}/compose"

            if repo == "extras":
                yield RpmRepository(
                    repo.lower(),
                    f"baseurl={join_mirror(mirror, f'{subdir}/{repo}/$basearch/extras-common')}",
                    gpgurls,
                )
                yield RpmRepository(
                    f"{repo.lower()}-source",
                    f"baseurl={join_mirror(mirror, f'{subdir}/{repo}/source/extras-common')}",
                    gpgurls,
                    enabled=False,
                )
            else:
                yield RpmRepository(
                    repo.lower(),
                    f"baseurl={join_mirror(mirror, f'{subdir}/{repo}/$basearch/os')}",
                    gpgurls,
                )
                yield RpmRepository(
                    f"{repo.lower()}-debuginfo",
                    f"baseurl={join_mirror(mirror, f'{subdir}/{repo}/$basearch/debug/tree')}",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo.lower()}-source",
                    f"baseurl={join_mirror(mirror, f'{subdir}/{repo}/source/tree')}",
                    gpgurls,
                    enabled=False,
                )
        else:
            url = "metalink=https://mirrors.centos.org/metalink"

            if repo == "extras":
                yield RpmRepository(
                    repo.lower(),
                    f"{url}?arch=$basearch&repo=centos-extras-sig-extras-common-$stream",
                    gpgurls,
                )
                yield RpmRepository(
                    f"{repo.lower()}-source",
                    f"{url}?arch=source&repo=centos-extras-sig-extras-common-source-$stream",
                    gpgurls,
                    enabled=False,
                )
            else:
                yield RpmRepository(
                    repo.lower(),
                    f"{url}?arch=$basearch&repo=centos-{repo.lower()}-$stream",
                    gpgurls,
                )
                yield RpmRepository(
                    f"{repo.lower()}-debuginfo",
                    f"{url}?arch=$basearch&repo=centos-{repo.lower()}-debug-$stream",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo.lower()}-source",
                    f"{url}?arch=source&repo=centos-{repo.lower()}-source-$stream",
                    gpgurls,
                    enabled=False,
                )

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        gpgurls = cls.gpgurls(context)

        if context.config.local_mirror:
            yield from cls.repository_variants(context, gpgurls, "AppStream")
            return

        yield from cls.repository_variants(context, gpgurls, "BaseOS")
        yield from cls.repository_variants(context, gpgurls, "AppStream")
        yield from cls.repository_variants(context, gpgurls, "CRB")
        if not context.config.snapshot:
            yield from cls.repository_variants(context, gpgurls, "extras")

        yield from cls.epel_repositories(context)
        yield from cls.sig_repositories(context)

    @classmethod
    def epel_repositories(cls, context: Context) -> Iterable[RpmRepository]:
        # Since EPEL 10, there's an associated minor release for every RHEL minor release.
        if GenericVersion(context.config.release) >= 10:
            release = context.config.release
        else:
            release = cls.major_release(context.config)

        gpgurls = (
            find_rpm_gpgkey(
                context,
                f"RPM-GPG-KEY-EPEL-{cls.major_release(context.config)}",
                f"https://dl.fedoraproject.org/pub/epel/RPM-GPG-KEY-EPEL-{cls.major_release(context.config)}",
            ),
        )

        if context.config.local_mirror:
            return

        if mirror := context.config.mirror:
            # epel-next does not exist anymore since EPEL 10.
            repodirs = [
                ("epel", "epel"),
                ("epel-testing", "epel/testing"),
            ]
            if GenericVersion(context.config.release) < 10:
                repodirs += [
                    ("epel-next", "epel/next"),
                    ("epel-next-testing", "epel/testing/next"),
                ]

            for repo, dir in repodirs:
                # For EPEL we make the assumption that epel is mirrored in the parent directory of the mirror
                # URL and path we were given. Since this doesn't work for all scenarios, we also allow
                # overriding the mirror via an environment variable.
                url = context.config.finalize_environment().get(
                    "EPEL_MIRROR", join_mirror(mirror, "../fedora")
                )
                yield RpmRepository(
                    repo,
                    f"baseurl={url}/{dir}/{release}/Everything/$basearch",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo}-debuginfo",
                    f"baseurl={url}/{dir}/{release}/Everything/$basearch/debug",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo}-source",
                    f"baseurl={url}/{dir}/{release}/Everything/source/tree",
                    gpgurls,
                    enabled=False,
                )
        else:
            url = "metalink=https://mirrors.fedoraproject.org/metalink?arch=$basearch"

            # epel-next does not exist anymore since EPEL 10.
            repos = ["epel"]
            if GenericVersion(context.config.release) < 10:
                repos += ["epel-next"]

            for repo in repos:
                yield RpmRepository(
                    repo,
                    f"{url}&repo={repo}-{release}",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo}-debuginfo",
                    f"{url}&repo={repo}-debug-{release}",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    f"{repo}-source",
                    f"{url}&repo={repo}-source-{release}",
                    gpgurls,
                    enabled=False,
                )

            yield RpmRepository(
                "epel-testing",
                f"{url}&repo=testing-epel{release}",
                gpgurls,
                enabled=False,
            )
            yield RpmRepository(
                "epel-testing-debuginfo",
                f"{url}&repo=testing-debug-epel{release}",
                gpgurls,
                enabled=False,
            )
            yield RpmRepository(
                "epel-testing-source",
                f"{url}&repo=testing-source-epel{release}",
                gpgurls,
                enabled=False,
            )

            # epel-next does not exist anymore since EPEL 10.
            if GenericVersion(context.config.release) < 10:
                yield RpmRepository(
                    "epel-next-testing",
                    f"{url}&repo=epel-testing-next-{release}",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    "epel-next-testing-debuginfo",
                    f"{url}&repo=epel-testing-next-debug-{release}",
                    gpgurls,
                    enabled=False,
                )
                yield RpmRepository(
                    "epel-next-testing-source",
                    f"{url}&repo=epel-testing-next-source-{release}",
                    gpgurls,
                    enabled=False,
                )

    @classmethod
    def sig_repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.local_mirror:
            return

        sigs = (
            (
                "hyperscale",
                (
                    f"packages-{c}"
                    for c in ("main", "experimental", "facebook", "hotfixes", "spin", "intel", "kernel")
                ),
                ("RPM-GPG-KEY-CentOS-SIG-HyperScale",),
            ),
        )

        for sig, components, keys in sigs:
            gpgurls = tuple(
                find_rpm_gpgkey(context, key, f"https://www.centos.org/keys/{key}") for key in keys
            )

            for c in components:
                if mirror := context.config.mirror:
                    yield RpmRepository(
                        f"{sig}-{c}",
                        f"baseurl={join_mirror(mirror, f'SIGs/$stream/{sig}/$basearch/{c}')}",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-debuginfo",
                        f"baseurl={join_mirror(mirror, f'SIGs/$stream/{sig}/$basearch/{c}/debug')}",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-source",
                        f"baseurl={join_mirror(mirror, f'SIGs/$stream/{sig}/source/{c}')}",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                else:
                    url = "metalink=https://mirrors.centos.org/metalink"
                    yield RpmRepository(
                        f"{sig}-{c}",
                        f"{url}?arch=$basearch&repo=centos-{sig}-sig-{c}-$stream",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-debuginfo",
                        f"{url}?arch=$basearch&repo=centos-{sig}-sig-{c}-debug-$stream",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-source",
                        f"{url}?arch=source&repo=centos-{sig}-sig-{c}-source-$stream",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-testing",
                        f"baseurl=https://buildlogs.centos.org/centos/$stream/{sig}/$basearch/{c}",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )
                    yield RpmRepository(
                        f"{sig}-{c}-testing-debuginfo",
                        f"baseurl=https://buildlogs.centos.org/centos/$stream/{sig}/$basearch/{c}",
                        gpgurls,
                        enabled=False,
                        priority=CENTOS_SIG_REPO_PRIORITY,
                    )

    @classmethod
    def latest_snapshot(cls, config: Config) -> str:
        mirror = config.mirror or "https://composes.stream.centos.org"

        if mirror == "https://mirror.facebook.net/centos-composes":
            subdir = config.release
        else:
            subdir = f"stream-{config.release}/production"

        url = join_mirror(mirror, f"{subdir}/latest-CentOS-Stream/compose/.composeinfo")

        for line in curl(config, url).splitlines():
            if snapshot := startswith(line, f"id = CentOS-Stream-{config.release}-"):
                return snapshot

        die("composeinfo is missing compose ID field")

    @classmethod
    def is_kernel_package(cls, package: str) -> bool:
        return package in ("kernel", "kernel-core")



# SPDX-License-Identifier: LGPL-2.1-or-later

import dataclasses
import shutil
from collections.abc import Sequence
from pathlib import Path

from mkosi.config import Config
from mkosi.context import Context
from mkosi.installer import PackageManager
from mkosi.run import CompletedProcess, run, workdir
from mkosi.tree import rmtree
from mkosi.util import _FILE, PathString


@dataclasses.dataclass(frozen=True)
class ApkRepository:
    url: str


class Apk(PackageManager):
    @classmethod
    def executable(cls, config: Config) -> str:
        return "apk"

    @classmethod
    def subdir(cls, config: Config) -> Path:
        return Path("apk")

    @classmethod
    def package_subdirs(cls, cache: Path) -> list[tuple[Path, Path]]:
        return [(Path("."), Path("."))]

    @classmethod
    def package_globs(cls) -> list[str]:
        return ["*.apk"]

    @classmethod
    def scripts(cls, context: Context) -> dict[str, list[PathString]]:
        return {
            "apk": cls.apivfs_script_cmd(context) + cls.env_cmd(context) + cls.cmd(context),
            "mkosi-install":   ["apk", "add", "--upgrade", "--cache-max-age", "999999999"],
            "mkosi-upgrade":   ["apk", "upgrade"],
            "mkosi-remove":    ["apk", "--remove", "del"],
            "mkosi-reinstall": ["apk", "fix", "--reinstall"],
        }  # fmt: skip

    @classmethod
    def setup(cls, context: Context, repositories: Sequence[ApkRepository]) -> None:
        config = context.sandbox_tree / "etc/apk/repositories"
        if config.exists():
            return

        config.parent.mkdir(exist_ok=True, parents=True)
        config.write_text("\n".join(repo.url for repo in repositories) + "\n")

    @classmethod
    def finalize_environment(cls, context: Context) -> dict[str, str]:
        return super().finalize_environment(context) | {
            # apk requires SHA1 support for signature verification, and this is disabled in the default
            # crypto-policies for Fedora/RH/SuSE. This variable is set to re-enable SHA1 support on these
            # distributions.
            # Also see: https://gitlab.alpinelinux.org/alpine/apk-tools/-/issues/11139#note_542183
            "OPENSSL_ENABLE_SHA1_SIGNATURES": "1",
        }

    @classmethod
    def cmd(cls, context: Context) -> list[PathString]:
        return [
            "apk",
            "--root", "/buildroot",
            "--cache-packages",
            "--cache-dir", "/var/cache/apk",
            "--cache-predownload",
            "--arch", cls.architecture(context),
            "--no-interactive",
            "--preserve-env",
            "--keys-dir", "/etc/apk/keys",
            "--repositories-file", "/etc/apk/repositories",
            *(["--allow-untrusted"] if not context.config.repository_key_check else []),
        ]  # fmt: skip

    @classmethod
    def invoke(
        cls,
        context: Context,
        operation: str,
        arguments: Sequence[str] = (),
        *,
        apivfs: bool = False,
        stdout: _FILE = None,
    ) -> CompletedProcess:
        return run(
            cls.cmd(context) + [operation, *arguments],
            sandbox=cls.sandbox(context, apivfs=apivfs),
            env=cls.finalize_environment(context),
            stdout=stdout,
        )

    @classmethod
    def install(
        cls,
        context: Context,
        packages: Sequence[str],
        *,
        apivfs: bool = True,
        allow_downgrade: bool = False,
    ) -> None:
        cls.invoke(
            context,
            "add",
            [
                "--initdb",
                "--upgrade",
                # effectively disable refreshing the cache in this situation
                "--cache-max-age", "999999999",
                *packages,
            ],
            apivfs=apivfs,
        )  # fmt: skip

    @classmethod
    def remove(cls, context: Context, packages: Sequence[str]) -> None:
        cls.invoke(context, "del", packages, apivfs=True)

    @classmethod
    def sync(cls, context: Context, force: bool) -> None:
        # Updating the cache requires an initialized apk database but we don't want to touch the image root
        # directory so temporarily replace it with an empty directory to make apk happy.
        saved = context.root.rename(context.workspace / "saved-root")
        context.root.mkdir()
        cls.invoke(context, "add", ["--initdb"])
        cls.invoke(context, "update", ["--update-cache"] if force else [])
        rmtree(context.root)
        saved.rename(context.root)

    @classmethod
    def createrepo(cls, context: Context) -> None:
        packages = [p.name for p in context.repository.glob("*.apk")]
        if not packages:
            return

        # Move apk files to arch-specific directory
        arch = cls.architecture(context)
        arch_dir = context.repository / arch
        arch_dir.mkdir(exist_ok=True)
        for package in packages:
            (context.repository / package).rename(arch_dir / package)

        # Generate temporary signing key using openssl
        # This uses the same method as abuild-keygen, because this tool is not available on all distros
        key = "mkosi@local"
        priv = context.workspace / f"{key}.rsa"
        pub = context.workspace / f"{key}.rsa.pub"

        if not priv.exists():
            run(
                ["openssl", "genrsa", "-out", workdir(priv), "2048"],
                sandbox=context.sandbox(options=["--bind", priv.parent, workdir(priv.parent)]),
                env=cls.finalize_environment(context),
            )
            run(
                ["openssl", "rsa", "-in", workdir(priv), "-pubout", "-out", workdir(pub)],
                sandbox=context.sandbox(
                    options=[
                        "--bind", priv, workdir(priv),
                        "--bind", pub.parent, workdir(pub.parent),
                    ]
                ),
                env=cls.finalize_environment(context),
            )  # fmt: skip
            keys = context.sandbox_tree / "etc/apk/keys"
            keys.mkdir(parents=True, exist_ok=True)
            shutil.copy2(pub, keys / pub.name)

        run(
            [
                "apk",
                "index",
                "-o", "APKINDEX.tar.gz",
                "--rewrite-arch", arch,
                # packages may be signed by another key that might not be available.
                "--allow-untrusted",
                *packages,
            ],
            sandbox=context.sandbox(
                options=[
                    "--bind", context.repository, workdir(context.repository),
                    "--chdir", workdir(arch_dir),
                ]
            ),
            env=cls.finalize_environment(context),
        )  # fmt: skip

        # Create and sign index signature file
        # Note: The index signing stuff below was largely inspired by what abuild-sign and abuild-tar tools
        # do on Alpine Linux. These tools are not always packages for other distros.
        index = arch_dir / "APKINDEX.tar.gz"
        sig = arch_dir / f".SIGN.RSA.{pub.name}"
        run(
            [
                "openssl",
                "dgst",
                "-sha1",
                "-sign", workdir(priv),
                "-out", workdir(sig),
                workdir(index),
            ],
            sandbox=context.sandbox(
                options=[
                    "--bind", priv, workdir(priv),
                    "--bind", sig.parent, workdir(sig.parent),
                    "--bind", index, workdir(index),
                ]
            ),
            env=cls.finalize_environment(context),
        )  # fmt: skip

        tar = context.workspace / "sig.tar"
        with tar.open("wb") as f:
            run(
                [
                    "tar", "-cf", "-",
                    "--format=posix",
                    "--owner=0",
                    "--group=0",
                    "--numeric-owner",
                    "-C", workdir(arch_dir),
                    sig.name,
                ],
                sandbox=context.sandbox(options=["--bind", arch_dir, workdir(arch_dir)]),
                stdout=f,
            )  # fmt: skip

        # Strip EOF markers to allow concatenation with compressed index.
        data = tar.read_bytes()
        while data.endswith(b"\x00" * 512):
            data = data[:-512]
        tar.write_bytes(data)

        # Prepend gzipped signature to original index

        signed = context.workspace / "signed.tar.gz"
        with signed.open("wb") as out:
            run(
                ["gzip", "-n", "-9", "-c", workdir(tar)],
                sandbox=context.sandbox(options=["--bind", tar, workdir(tar)]),
                stdout=out,
            )
            out.write(index.read_bytes())

        # Finally, overwrite the original index archive with the signed index archive
        signed.rename(index)

        repos = context.sandbox_tree / "etc/apk/repositories"
        repo = "file:///repository/"
        if repos.exists():
            content = repos.read_text()
            if repo not in content:
                with repos.open("a") as f:
                    f.write(f"{repo}\n")
        else:
            repos.write_text(f"{repo}\n")

        cls.sync(context, force=True)



# SPDX-License-Identifier: LGPL-2.1-or-later

from collections.abc import Iterable

from mkosi.context import Context
from mkosi.distribution import Distribution, centos, join_mirror
from mkosi.installer.rpm import RpmRepository, find_rpm_gpgkey
from mkosi.log import die


class Installer(centos.Installer, distribution=Distribution.rhel_ubi):
    @classmethod
    def pretty_name(cls) -> str:
        return "RHEL UBI"

    @classmethod
    def gpgurls(cls, context: Context) -> tuple[str, ...]:
        return (
            find_rpm_gpgkey(
                context,
                f"RPM-GPG-KEY-redhat{cls.major_release(context.config)}-release",
                "https://access.redhat.com/security/data/fd431d51.txt",
            ),
        )

    @classmethod
    def repository_variants(
        cls,
        context: Context,
        gpgurls: tuple[str, ...],
        repo: str,
    ) -> Iterable[RpmRepository]:
        if context.config.local_mirror:
            yield RpmRepository(repo, f"baseurl={context.config.local_mirror}", gpgurls)
        else:
            mirror = context.config.mirror or "https://cdn-ubi.redhat.com/content/public/ubi/dist/"

            v = context.config.release
            yield RpmRepository(
                f"ubi-{v}-{repo}-rpms",
                f"baseurl={join_mirror(mirror, f'ubi{v}/{v}/$basearch/{repo}/os')}",
                gpgurls,
            )
            yield RpmRepository(
                f"ubi-{v}-{repo}-debug-rpms",
                f"baseurl={join_mirror(mirror, f'ubi{v}/{v}/$basearch/{repo}/debug')}",
                gpgurls,
                enabled=False,
            )
            yield RpmRepository(
                f"ubi-{v}-{repo}-source",
                f"baseurl={join_mirror(mirror, f'ubi{v}/{v}/$basearch/{repo}/source')}",
                gpgurls,
                enabled=False,
            )

    @classmethod
    def repositories(cls, context: Context) -> Iterable[RpmRepository]:
        if context.config.snapshot:
            die(f"Snapshot= is not supported for {cls.pretty_name()}")

        gpgurls = cls.gpgurls(context)
        yield from cls.repository_variants(context, gpgurls, "baseos")
        yield from cls.repository_variants(context, gpgurls, "appstream")
        yield from cls.repository_variants(context, gpgurls, "codeready-builder")
        yield from cls.epel_repositories(context)
